"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var isMaybePropertyName_1 = require("./util/isMaybePropertyName");
/**
 * Class responsible for preprocessing and bookkeeping import and export
 * declarations within the file.
 */
var ImportProcessor = /** @class */ (function () {
    function ImportProcessor(nameManager, tokens) {
        this.nameManager = nameManager;
        this.tokens = tokens;
        this.importInfoByPath = new Map();
        this.importsToReplace = new Map();
        this.identifierReplacements = new Map();
        this.exportBindingsByLocalName = new Map();
    }
    ImportProcessor.prototype.getPrefixCode = function () {
        var prefix = '';
        prefix += ("\n      function " + this.interopRequireWildcardName + "(obj) {\n        if (obj && obj.__esModule) {\n          return obj;\n        } else {\n          var newObj = {};\n          if (obj != null) {\n            for (var key in obj) {\n              if (Object.prototype.hasOwnProperty.call(obj, key))\n                newObj[key] = obj[key];\n              }\n            }\n          newObj.default = obj;\n          return newObj;\n        }\n      }").replace(/\s+/g, ' ');
        prefix += ("\n      function " + this.interopRequireDefaultName + "(obj) {\n        return obj && obj.__esModule ? obj : { default: obj };\n      }").replace(/\s+/g, ' ');
        return prefix;
    };
    ImportProcessor.prototype.preprocessTokens = function () {
        this.interopRequireWildcardName = this.nameManager.claimFreeName('_interopRequireWildcard');
        this.interopRequireDefaultName = this.nameManager.claimFreeName('_interopRequireDefault');
        for (var i = 0; i < this.tokens.tokens.length; i++) {
            if (isMaybePropertyName_1.default(this.tokens, i)) {
                continue;
            }
            if (this.tokens.matchesAtIndex(i, ['import'])) {
                this.preprocessImportAtIndex(i);
            }
            if (this.tokens.matchesAtIndex(i, ['export'])) {
                this.preprocessExportAtIndex(i);
            }
        }
        this.generateImportReplacements();
    };
    ImportProcessor.prototype.generateImportReplacements = function () {
        try {
            for (var _a = tslib_1.__values(this.importInfoByPath.entries()), _b = _a.next(); !_b.done; _b = _a.next()) {
                var _c = tslib_1.__read(_b.value, 2), path = _c[0], importInfo = _c[1];
                var defaultNames = importInfo.defaultNames, wildcardNames = importInfo.wildcardNames, namedImports = importInfo.namedImports, namedExports = importInfo.namedExports, hasStarExport = importInfo.hasStarExport;
                if (defaultNames.length === 0 && wildcardNames.length === 0 && namedImports.length === 0 && namedExports.length === 0 && !hasStarExport) {
                    // Import is never used, so don't even assign a name.
                    this.importsToReplace.set(path, "require('" + path + "');");
                    continue;
                }
                var primaryImportName = this.getFreeIdentifierForPath(path);
                var secondaryImportName = wildcardNames.length > 0
                    ? wildcardNames[0]
                    : this.getFreeIdentifierForPath(path);
                var requireCode = "var " + primaryImportName + " = require('" + path + "');";
                if (wildcardNames.length > 0) {
                    try {
                        for (var wildcardNames_1 = tslib_1.__values(wildcardNames), wildcardNames_1_1 = wildcardNames_1.next(); !wildcardNames_1_1.done; wildcardNames_1_1 = wildcardNames_1.next()) {
                            var wildcardName = wildcardNames_1_1.value;
                            requireCode += " var " + wildcardName + " = " + this.interopRequireWildcardName + "(" + primaryImportName + ");";
                        }
                    }
                    catch (e_1_1) { e_1 = { error: e_1_1 }; }
                    finally {
                        try {
                            if (wildcardNames_1_1 && !wildcardNames_1_1.done && (_d = wildcardNames_1.return)) _d.call(wildcardNames_1);
                        }
                        finally { if (e_1) throw e_1.error; }
                    }
                }
                else if (defaultNames.length > 0) {
                    requireCode += " var " + secondaryImportName + " = " + this.interopRequireDefaultName + "(" + primaryImportName + ");";
                }
                try {
                    for (var namedExports_1 = tslib_1.__values(namedExports), namedExports_1_1 = namedExports_1.next(); !namedExports_1_1.done; namedExports_1_1 = namedExports_1.next()) {
                        var _e = namedExports_1_1.value, importedName = _e.importedName, localName = _e.localName;
                        requireCode += " Object.defineProperty(exports, '" + localName + "', {enumerable: true, get: () => " + primaryImportName + "." + importedName + "});";
                    }
                }
                catch (e_2_1) { e_2 = { error: e_2_1 }; }
                finally {
                    try {
                        if (namedExports_1_1 && !namedExports_1_1.done && (_f = namedExports_1.return)) _f.call(namedExports_1);
                    }
                    finally { if (e_2) throw e_2.error; }
                }
                if (hasStarExport) {
                    requireCode += " Object.keys(" + primaryImportName + ").filter(key => key !== 'default' && key !== '__esModule').forEach(key => { Object.defineProperty(exports, key, {enumerable: true, get: () => " + primaryImportName + "[key]}); });";
                }
                this.importsToReplace.set(path, requireCode);
                try {
                    for (var defaultNames_1 = tslib_1.__values(defaultNames), defaultNames_1_1 = defaultNames_1.next(); !defaultNames_1_1.done; defaultNames_1_1 = defaultNames_1.next()) {
                        var defaultName = defaultNames_1_1.value;
                        this.identifierReplacements.set(defaultName, secondaryImportName + ".default");
                    }
                }
                catch (e_3_1) { e_3 = { error: e_3_1 }; }
                finally {
                    try {
                        if (defaultNames_1_1 && !defaultNames_1_1.done && (_g = defaultNames_1.return)) _g.call(defaultNames_1);
                    }
                    finally { if (e_3) throw e_3.error; }
                }
                try {
                    for (var namedImports_1 = tslib_1.__values(namedImports), namedImports_1_1 = namedImports_1.next(); !namedImports_1_1.done; namedImports_1_1 = namedImports_1.next()) {
                        var _h = namedImports_1_1.value, importedName = _h.importedName, localName = _h.localName;
                        this.identifierReplacements.set(localName, primaryImportName + "." + importedName);
                    }
                }
                catch (e_4_1) { e_4 = { error: e_4_1 }; }
                finally {
                    try {
                        if (namedImports_1_1 && !namedImports_1_1.done && (_j = namedImports_1.return)) _j.call(namedImports_1);
                    }
                    finally { if (e_4) throw e_4.error; }
                }
            }
        }
        catch (e_5_1) { e_5 = { error: e_5_1 }; }
        finally {
            try {
                if (_b && !_b.done && (_k = _a.return)) _k.call(_a);
            }
            finally { if (e_5) throw e_5.error; }
        }
        var e_5, _k, e_1, _d, e_2, _f, e_3, _g, e_4, _j;
    };
    ImportProcessor.prototype.getFreeIdentifierForPath = function (path) {
        var components = path.split('/');
        var lastComponent = components[components.length - 1];
        var baseName = lastComponent.replace(/\W/g, '');
        return this.nameManager.claimFreeName("_" + baseName);
    };
    ImportProcessor.prototype.preprocessImportAtIndex = function (index) {
        var defaultNames = [];
        var wildcardNames = [];
        var namedImports = [];
        index++;
        if (this.tokens.matchesAtIndex(index, ['name'])) {
            defaultNames.push(this.tokens.tokens[index].value);
            index++;
            if (this.tokens.matchesAtIndex(index, [','])) {
                index++;
            }
        }
        if (this.tokens.matchesAtIndex(index, ['*'])) {
            // * as
            index += 2;
            wildcardNames.push(this.tokens.tokens[index].value);
            index++;
        }
        if (this.tokens.matchesAtIndex(index, ['{'])) {
            index++;
            (_a = this.getNamedImports(index), index = _a.newIndex, namedImports = _a.namedImports);
        }
        if (this.tokens.matchesNameAtIndex(index, 'from')) {
            index++;
        }
        if (!this.tokens.matchesAtIndex(index, ['string'])) {
            throw new Error('Expected string token at the end of import statement.');
        }
        var path = this.tokens.tokens[index].value;
        var importInfo = this.getImportInfo(path);
        (_b = importInfo.defaultNames).push.apply(_b, tslib_1.__spread(defaultNames));
        (_c = importInfo.wildcardNames).push.apply(_c, tslib_1.__spread(wildcardNames));
        (_d = importInfo.namedImports).push.apply(_d, tslib_1.__spread(namedImports));
        var _a, _b, _c, _d;
    };
    ImportProcessor.prototype.preprocessExportAtIndex = function (index) {
        if (this.tokens.matchesAtIndex(index, ['export', 'var']) ||
            this.tokens.matchesAtIndex(index, ['export', 'let']) ||
            this.tokens.matchesAtIndex(index, ['export', 'const']) ||
            this.tokens.matchesAtIndex(index, ['export', 'function']) ||
            this.tokens.matchesAtIndex(index, ['export', 'class'])) {
            var exportName = this.tokens.tokens[index + 2].value;
            this.exportBindingsByLocalName.set(exportName, exportName);
        }
        else if (this.tokens.matchesAtIndex(index, ['export', 'name', 'function'])) {
            var exportName = this.tokens.tokens[index + 3].value;
            this.exportBindingsByLocalName.set(exportName, exportName);
        }
        else if (this.tokens.matchesAtIndex(index, ['export', '{'])) {
            this.preprocessNamedExportAtIndex(index);
        }
        else if (this.tokens.matchesAtIndex(index, ['export', '*'])) {
            this.preprocessExportStarAtIndex(index);
        }
    };
    /**
     * Walk this export statement just in case it's an export...from statement.
     * If it is, combine it into the import info for that path. Otherwise, just
     * bail out; it'll be handled later.
     */
    ImportProcessor.prototype.preprocessNamedExportAtIndex = function (index) {
        // export {
        index += 2;
        var _a = this.getNamedImports(index), newIndex = _a.newIndex, namedImports = _a.namedImports;
        index = newIndex;
        if (this.tokens.matchesNameAtIndex(index, 'from')) {
            index++;
        }
        else {
            try {
                // Reinterpret "a as b" to be local/exported rather than imported/local.
                for (var namedImports_2 = tslib_1.__values(namedImports), namedImports_2_1 = namedImports_2.next(); !namedImports_2_1.done; namedImports_2_1 = namedImports_2.next()) {
                    var _b = namedImports_2_1.value, localName = _b.importedName, exportedName = _b.localName;
                    this.exportBindingsByLocalName.set(localName, exportedName);
                }
            }
            catch (e_6_1) { e_6 = { error: e_6_1 }; }
            finally {
                try {
                    if (namedImports_2_1 && !namedImports_2_1.done && (_c = namedImports_2.return)) _c.call(namedImports_2);
                }
                finally { if (e_6) throw e_6.error; }
            }
            return;
        }
        if (!this.tokens.matchesAtIndex(index, ['string'])) {
            throw new Error('Expected string token at the end of import statement.');
        }
        var path = this.tokens.tokens[index].value;
        var importInfo = this.getImportInfo(path);
        (_d = importInfo.namedExports).push.apply(_d, tslib_1.__spread(namedImports));
        var e_6, _c, _d;
    };
    ImportProcessor.prototype.preprocessExportStarAtIndex = function (index) {
        // export * from
        index += 3;
        if (!this.tokens.matchesAtIndex(index, ['string'])) {
            throw new Error('Expected string token at the end of star export statement.');
        }
        var path = this.tokens.tokens[index].value;
        var importInfo = this.getImportInfo(path);
        importInfo.hasStarExport = true;
    };
    ImportProcessor.prototype.getNamedImports = function (index) {
        var namedImports = [];
        while (true) {
            var importedName = this.tokens.tokens[index].value;
            var localName = void 0;
            index++;
            if (this.tokens.matchesNameAtIndex(index, 'as')) {
                index++;
                localName = this.tokens.tokens[index].value;
                index++;
            }
            else {
                localName = importedName;
            }
            namedImports.push({ importedName: importedName, localName: localName });
            if (this.tokens.matchesAtIndex(index, [',', '}'])) {
                index += 2;
                break;
            }
            else if (this.tokens.matchesAtIndex(index, ['}'])) {
                index++;
                break;
            }
            else if (this.tokens.matchesAtIndex(index, [','])) {
                index++;
            }
            else {
                throw new Error('Unexpected token.');
            }
        }
        return { newIndex: index, namedImports: namedImports };
    };
    /**
     * Get a mutable import info object for this path, creating one if it doesn't
     * exist yet.
     */
    ImportProcessor.prototype.getImportInfo = function (path) {
        var existingInfo = this.importInfoByPath.get(path);
        if (existingInfo) {
            return existingInfo;
        }
        else {
            var newInfo = {
                defaultNames: [],
                wildcardNames: [],
                namedImports: [],
                namedExports: [],
                hasStarExport: false,
            };
            this.importInfoByPath.set(path, newInfo);
            return newInfo;
        }
    };
    /**
     * Return the code to use for the import for this path, or the empty string if
     * the code has already been "claimed" by a previous import.
     */
    ImportProcessor.prototype.claimImportCode = function (importPath) {
        var result = this.importsToReplace.get(importPath);
        this.importsToReplace.set(importPath, '');
        return result || '';
    };
    ImportProcessor.prototype.getIdentifierReplacement = function (identifierName) {
        return this.identifierReplacements.get(identifierName) || null;
    };
    ImportProcessor.prototype.resolveExportBinding = function (assignedName) {
        return this.exportBindingsByLocalName.get(assignedName) || null;
    };
    return ImportProcessor;
}());
exports.ImportProcessor = ImportProcessor;
