import * as tslib_1 from "tslib";
import { parse } from 'babylon';
import JSXTransformer from './transformers/JSXTransformer';
import TokenProcessor from './TokenProcessor';
import ImportTransformer from './transformers/ImportTransformer';
import augmentTokenContext from './augmentTokenContext';
import { ImportProcessor } from './ImportProcessor';
import { NameManager } from './NameManager';
import ReactDisplayNameTransformer from './transformers/ReactDisplayNameTransformer';
var DEFAULT_BABYLON_PLUGINS = ['jsx', 'objectRestSpread'];
export function transform(code, options) {
    if (options === void 0) { options = {}; }
    var babylonPlugins = options.babylonPlugins || DEFAULT_BABYLON_PLUGINS;
    var transforms = options.transforms || ['jsx'];
    var ast = parse(code, { tokens: true, sourceType: 'module', plugins: babylonPlugins });
    var tokens = ast.tokens;
    tokens = tokens.filter(function (token) {
        return token.type !== 'CommentLine' && token.type !== 'CommentBlock';
    });
    augmentTokenContext(tokens);
    var tokenProcessor = new TokenProcessor(code, tokens);
    return new RootTransformer(tokenProcessor, transforms).transform();
}
var RootTransformer = /** @class */ (function () {
    function RootTransformer(tokens, transforms) {
        this.tokens = tokens;
        this.transformers = [];
        var nameManager = new NameManager(tokens);
        var importProcessor = transforms.includes('imports')
            ? new ImportProcessor(nameManager, tokens)
            : null;
        var identifierReplacer = importProcessor
            ? importProcessor
            : { getIdentifierReplacement: function () { return null; } };
        if (transforms.includes('jsx')) {
            this.transformers.push(new JSXTransformer(this, tokens, identifierReplacer));
        }
        // react-display-name must come before imports since otherwise imports will
        // claim a normal `React` name token.
        if (transforms.includes('react-display-name')) {
            this.transformers.push(new ReactDisplayNameTransformer(this, tokens, identifierReplacer));
        }
        if (transforms.includes('imports')) {
            var shouldAddModuleExports = transforms.includes('add-module-exports');
            this.transformers.push(new ImportTransformer(this, tokens, nameManager, importProcessor, shouldAddModuleExports));
        }
    }
    RootTransformer.prototype.transform = function () {
        this.tokens.reset();
        try {
            for (var _a = tslib_1.__values(this.transformers), _b = _a.next(); !_b.done; _b = _a.next()) {
                var transformer = _b.value;
                transformer.preprocess();
            }
        }
        catch (e_1_1) { e_1 = { error: e_1_1 }; }
        finally {
            try {
                if (_b && !_b.done && (_c = _a.return)) _c.call(_a);
            }
            finally { if (e_1) throw e_1.error; }
        }
        this.processBalancedCode();
        var prefix = '';
        try {
            for (var _d = tslib_1.__values(this.transformers), _e = _d.next(); !_e.done; _e = _d.next()) {
                var transformer = _e.value;
                prefix += transformer.getPrefixCode();
            }
        }
        catch (e_2_1) { e_2 = { error: e_2_1 }; }
        finally {
            try {
                if (_e && !_e.done && (_f = _d.return)) _f.call(_d);
            }
            finally { if (e_2) throw e_2.error; }
        }
        var suffix = '';
        try {
            for (var _g = tslib_1.__values(this.transformers), _h = _g.next(); !_h.done; _h = _g.next()) {
                var transformer = _h.value;
                suffix += transformer.getSuffixCode();
            }
        }
        catch (e_3_1) { e_3 = { error: e_3_1 }; }
        finally {
            try {
                if (_h && !_h.done && (_j = _g.return)) _j.call(_g);
            }
            finally { if (e_3) throw e_3.error; }
        }
        return prefix + this.tokens.finish() + suffix;
        var e_1, _c, e_2, _f, e_3, _j;
    };
    RootTransformer.prototype.processBalancedCode = function () {
        var braceDepth = 0;
        var parenDepth = 0;
        while (!this.tokens.isAtEnd()) {
            var wasProcessed = false;
            try {
                for (var _a = tslib_1.__values(this.transformers), _b = _a.next(); !_b.done; _b = _a.next()) {
                    var transformer = _b.value;
                    wasProcessed = transformer.process();
                    if (wasProcessed) {
                        break;
                    }
                }
            }
            catch (e_4_1) { e_4 = { error: e_4_1 }; }
            finally {
                try {
                    if (_b && !_b.done && (_c = _a.return)) _c.call(_a);
                }
                finally { if (e_4) throw e_4.error; }
            }
            if (!wasProcessed) {
                if (this.tokens.matches(['{']) || this.tokens.matches(['${'])) {
                    braceDepth++;
                }
                else if (this.tokens.matches(['}'])) {
                    if (braceDepth === 0) {
                        return;
                    }
                    braceDepth--;
                }
                if (this.tokens.matches(['('])) {
                    parenDepth++;
                }
                else if (this.tokens.matches([')'])) {
                    if (parenDepth === 0) {
                        return;
                    }
                    parenDepth--;
                }
                this.tokens.copyToken();
            }
        }
        var e_4, _c;
    };
    RootTransformer.prototype.processToken = function () {
        try {
            for (var _a = tslib_1.__values(this.transformers), _b = _a.next(); !_b.done; _b = _a.next()) {
                var transformer = _b.value;
                var wasProcessed = transformer.process();
                if (wasProcessed) {
                    return;
                }
            }
        }
        catch (e_5_1) { e_5 = { error: e_5_1 }; }
        finally {
            try {
                if (_b && !_b.done && (_c = _a.return)) _c.call(_a);
            }
            finally { if (e_5) throw e_5.error; }
        }
        this.tokens.copyToken();
        var e_5, _c;
    };
    return RootTransformer;
}());
export { RootTransformer };
